/* 
 * Autor: Franz Brauße
 * License: GPL v2 
 */

package views;

import general.Config;

import javax.swing.ImageIcon;

import yacy.Peer;

/**
 * A plugin shall add new views and possibilities to administrate YaCy to YacyAdmin.
 */
public interface PanelPlugin extends Plugin {
	
	public static final String ADMIN = "Admin";
	public static final String CRAWL = "Crawl";
	public static final String STATUS = "Status";
	public static final String NETWORK = "Netzwerk";
	public static final String OTHER = "Other";
	
	/**
	 * Preset menu categories for plugins are used to hold confusion for the user and the peer
	 * menu tree small. It is not desirable to have a seperate individual category for every
	 * plugin. But you may mail me if an important plugin category is missing, I will add it
	 * if it makes sense.
	 */
	public static final String[] menuCategories = new String[] { ADMIN, CRAWL, STATUS, NETWORK, OTHER };
	
	/**
	 * Do final initializations. This function is called directly after instantiation.
	 * 
	 * @param peer The Peer-Object which offers all currently accessible data from the
	 * connected YaCy peer. Save it privately to be able to access the peer while the
	 * plugin is active.
	 */
	public void init(Peer peer, Config cfg);
	
	/**
	 * The description is the String which will be displayed in the left view
	 * and identifies this item. Therefore it must be unique.
	 * 
	 * @return The name of this plugin. 
	 */
	public String getDescription();
	
	/**
	 * The menu item enabling access to the plugin must be placed somewhere in the
	 * root menu. Choose the category, the plugin fits in best from menuCategories.
	 * Non-mentioned values will be rejected and the menu will be placed below "other" instead.
	 * 
	 * @return the name of the root node under which the plugin-menu entry shall be placed.
	 */
	public String getMenuPath();
	
	/**
	 * The menu icon is displayed in the tree on the left side of the main window. It shall
	 * not exceed the size of 16x16 pixel.
	 * 
	 * @return the menu icon representing the function of this plugin as {@link javax.swing#ImageIcon ImageIcon}
	 */
	public ImageIcon getMenuIcon();
	
	/**
	 * The menu tool-tip is displayed when the user hovers over the menu element in the tree
	 * on the left side of the main window. It shall give him an idea about this plugin.
	 * 
	 * @return a tool-tip text explaining shortly what this plugin is about and/or which
	 * functions it offers.
	 */
	public String getMenuToolTip();
	
	/**
	 * @return Time in milli seconds the plugin desires to update itself. Return 0 if no update
	 * is needed (due to i.e. no dynamic content). Set it to -1 to accept the programs default
	 * value This is the preferred behaviour if no time-critical information is needed, since
	 * it allows the user to control the bandwidth usage much better.
	 * If the user explicitely sets a different update interval in the program settings, this
	 * value will be overridden.
	 * <p>The default value is 3000 ms.</p>
	 */
	public int getUpdateDelay();
	
	/**
	 * Update all Components by using the Peer-Object received by the {@link #init(Peer, Config) init}-method call.
	 * 
	 * @deprecated use Peer-Events instead
	 * @see yacy.event for all currently available {@link yacy#Peer Peer}-Events
	 * (add / request new events if needed)
	 */
	public void doUpdate();
	
}
