/* 
 * Autor: Franz Brauße
 * License: GPL v2 
 */

package yacy;

import general.Config;

import yacy.datatypes.*;
import yacy.event.*;
import yacy.exceptions.*;
import yacy.soap.admin.*;
import yacy.soap.blacklist.*;
import yacy.soap.bookmarks.*;
import yacy.soap.crawl.*;
import yacy.soap.messages.*;
import yacy.soap.search.*;
import yacy.soap.share.*;
import yacy.soap.status.*;

import views.PanelPlugin;

import org.apache.axis.AxisFault;
import org.apache.axis.MessageContext;
import org.apache.axis.attachments.AttachmentPart;
import org.apache.axis.client.Call;
import org.apache.axis.client.Stub;
import org.apache.axis.transport.http.HTTPConstants;

import org.jdom.input.DOMBuilder;
import org.jdom.DataConversionException;
import org.jdom.Document;
import org.jdom.Element;

import java.io.File;
import java.io.IOException;
import java.rmi.RemoteException;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.List;
import java.util.ArrayList;
import java.util.HashMap;

import javax.activation.DataHandler;
import javax.activation.FileDataSource;
import javax.swing.event.EventListenerList;
import javax.xml.rpc.ServiceException;

/**
 * This class represents a YaCy peer.
 * It offers all functions currently implemented by the YaCy SOAP API by Theli 
 */
public class Peer {
	
	public static final int PEERS_ACTIVE 		= 0;
	public static final int PEERS_PASSIVE 		= 1;
	public static final int PEEERS_JUNIOR 		= 2;
	public static final int PEERS_ALL 			= 3;
	private static final String[] PEER_STATI = new String[] {
			"active", "passive", "junior", "all" };
	
	public static final String STATE_VIRGIN 	= "virgin";
	public static final String STATE_JUNIOR 	= "junior";
	public static final String STATE_SENIOR 	= "senior";
	public static final String STATE_PRINCIPAL 	= "principal";
	
	public static final String BLACKLIST_ENGINE_YACY = "de.anomic.plasma.UrlPattern.defaultUrlPattern";
	
	public static final int QUEUE_INDEXER 		= 0;
	public static final int QUEUE_LOADER 		= 1;
	public static final int QUEUE_LOCAL 		= 2;
	public static final int QUEUE_REMOTE 		= 3;
	public static final String[] QUEUE_TYPES = new String[] {
			"indexingqueue", "loaderqueue", "localcrawlerqueue", "remotecrawlerqueue" };
	
	// PeerSettingChangedEvent status messages
	public static final String SC_ADDED_BLACKLIST_ENTRY_TO 			= "added blacklist entry to %s";
	public static final String SC_ADDED_BOOKMARK 					= "added bookmark";
	public static final String SC_CREATED_SHARE_DIR 				= "created share directory";
	public static final String SC_DELETED_BOOKMARK 					= "deleted bookmark";
	public static final String SC_DELETED_BOOKMARKS 				= "deleted bookmarks";
	public static final String SC_DELETED_MESSAGE 					= "deleted message";
	public static final String SC_DELETED_MESSAGES 					= "deleted messages";
	public static final String SC_DELETED_SHARE_FILE 				= "deleted share file";
	public static final String SC_EDITED_BOOKMARK 					= "edited bookmark";
	public static final String SC_EDITED_SHARE_FILE_COMMENT 		= "edited share file comment of %s";
	public static final String SC_REMOVED_BLACKLIST_ITEM_FROM 		= "removed blacklist item from %s";
	public static final String SC_REMOVED_BLACKLIST 				= "removed blacklist";
	public static final String SC_RENAMED_BOOKMARK_TAG 				= "renamed bookmark tag";
	public static final String SC_SHARING_BLACKLIST 				= "%sshared blacklist %s ";
	public static final String SC_QUEUE_STATES_ALL 					= "all queue states changed";
	public static final String SC_INDEX_DISTRIBUTION 				= "changed index distribution";
	public static final String SC_INDEX_DISTRIBUTION_WHILE_CRAWLING = "changed index distribution while crawling";
	public static final String SC_INDEX_RECEIVE 					= "changed index receive";
	public static final String SC_INDEX_RECEIVE_BLOCKED 			= "changed index receive block blacklist";
	public static final String SC_PEER_NAME 						= "changed peer name";
	public static final String SC_PORT 								= "changed port";
	public static final String SC_INDEX_TRANSFER 					= "changed index transfer properties";
	public static final String SC_UPDATE_PROPERTIES 				= "changed update properties";
	
	// Logging output
	public static final String INFO_SVN_TOO_LOW_OVERRIDDEN 			=
		"Overridden by command-line, don't complain about errors...";
	
	public static final String WAR_SVN_TOO_LOW 						=
		"Warnung: Peer %s l\u00E4uft mit SVN %d.%nZum ordnungsgem\u00E4\u00DFen wird mindestens SVN %d vorrausgesetzt.";
	
	public static final String ERR_CONNECTION_RESET 				= "Connection reset";
	public static final String ERR_CONNECTION_REFUSED 				= "Connection refused";
	public static final String ERR_SERVICE_UNAVAILABLE 				= "Service unavailable";
	public static final String ERR_NOT_IMPLEMENTED 					= "Not Implemented";
	public static final String ERR_LOGIN_REQUIRED 					= "log-in required";
	
	// protocol-related strings
	private static final String ANOMIC_HEADER 	= "http://http.anomic.de/header";
	private static final String AUTHORIZATION 	= "Authorization";
	private static final String PROTOCOL_HTTP 	= "http://";
	private static final String PROTOCOL_HTTPS 	= "https://";
	private static final String SOAP_PATH 		= "/soap";
	private static final String ADMIN_PATH 		= "/admin";
	private static final String BLACKLIST_PATH 	= "/blacklist";
	private static final String BOOKMARKS_PATH 	= "/bookmarks";
	private static final String CRAWL_PATH 		= "/crawl";
	private static final String MESSAGES_PATH 	= "/messages";
	private static final String SEARCH_PATH 	= "/search";
	private static final String SHARE_PATH 		= "/share";
	private static final String STATUS_PATH 	= "/status";
	
	public static final String DEFAULT_USER = "admin";
	
	private String 	host = null;
	private int 	port = 8080;
	private String 	user = null;
	private String 	pwdEnc = null;
	private String 	pHost = null;
													// private static final int NONE 		= 0;
	private AdminService 		admin 		= null;	// private static final int ADMIN 		= 1;
	private BlacklistService 	blacklist 	= null;	// private static final int BLACKLIST 	= 2;
	private BookmarkService 	bookmark 	= null;	// private static final int BOOKMARKS 	= 4;
	private CrawlService 		crawl 		= null;	// private static final int CRAWL 		= 8;
	private MessageService 		message 	= null;	// private static final int MESSAGE 	= 16;
	private SearchService 		search 		= null; // private static final int SEARCH 		= 32;
	private ShareService 		share 		= null;	// private static final int SHARE 		= 64;
	private StatusService 		status 		= null;	// private static final int STATUS 		= 128;
	
	private String name;
	
	private List<Element> blist = null;
	private Document network = null;
	// private Document properties = null;
	
	private boolean indexDistributionEnabled 		= false;
	private boolean indexDistributionWhileCrawling 	= false;
	private boolean indexReceiveEnabled 			= false;
	private boolean indexReceiveBlockBlacklist 		= false;
	
	private long fromLogSequence = 0;
	private String peerHash = null;
	/*
	private Document transferProps = null;
	private Document queues = null;
	private int[] queueCounts = new int[] {5,5,5,5};
	*/
	
	private boolean useUserHeaderTable = false;
	protected static final Hashtable<String,String> userHeaderTable = new Hashtable<String,String>();
	static {
		userHeaderTable.put(HTTPConstants.HEADER_TRANSFER_ENCODING, HTTPConstants.HEADER_TRANSFER_ENCODING_CHUNKED);
	}
	
	private UpdateThread updateThread = null;
	private final Config cfg;
	private boolean connected = false;
	
	/* non-changing values */
	private long 		memoryMax 		= 0;
	private String 		blacklistEngine = BLACKLIST_ENGINE_YACY;
	private int 		svnRevision 	= 0;
	private String 		version 		= null;
	private String[] 	blacklistTypes 	= null;
	
	private final EventListenerList peerEventListeners = new EventListenerList();
	
	/* -----------------------------------------------------------------------
	 * -------------------------- admin functions ----------------------------
	 * ----------------------------------------------------------------------- */
	
	public Peer(Config cfg) {
		this.cfg = cfg;
		this.useUserHeaderTable = cfg.useChunkedTransfer;
	}
	
	/**
	 * @param host The publically accessable hostname of the YaCy peer 
	 * @param port The port on which it runs
	 * @param user
	 * @param passwordMD5Hex The {@link Codings.encodeMD5Hex}-encoded String conforming the following
	 * pattern: "user:password", where user normally is "admin" and password is the password
	 * for the specified user.
	 * @throws <code>RemoteException</code> if a general yet unknown connection error occured
	 * @throws <code>PLoginWrongEx</code> if username / password don't match
	 * @throws <code>PSVNRevisionTooLowEx</code> if the SOAP-backend on the peer doesn't contain all referenced methods 
	 * @throws <code>PConnectionRefusedEx</code> if the host refused the connection, i.e. <code>port</code> is wrong or YaCy isn't running
	 * @throws <code>PServiceUnavailableEx</code> if YaCy was compiled without SOAP
	 * @throws <code>PConnectionResetEx</code> if the host resetted the connection
	 */
	public void connect(String host, int port, String user, String passwordMD5Hex)
			throws RemoteException, PLoginWrongEx, PSVNRevisionTooLowEx, PConnectionRefusedEx,
			PServiceUnavailableEx, PConnectionResetEx {
		if (this.connected) return;
		if (host == null || host.length() == 0 || port < 1 || port > 65535 || user == null || user.length() == 0 || passwordMD5Hex == null || passwordMD5Hex.length() == 0)
			throw new NullPointerException(String.format("neither arguments must be null. (host: %s, port: %d, user: %s, pass: %s", host, port, user, passwordMD5Hex));
		this.host = host;
		this.port = port;
		this.user = user;
		this.pwdEnc = passwordMD5Hex;
		this.pHost = (new StringBuffer(20).append((cfg.useHTTPS) ? PROTOCOL_HTTPS : PROTOCOL_HTTP).append(host)
				.append(':').append(Integer.toString(port)).append(SOAP_PATH)).toString();
		fetchInitialValues();
	}
	
	/**
	 * Creates a new Peer-Object and connects to the peer immediately
	 * 
	 * @param host The publically accessable hostname of the YaCy peer 
	 * @param port The port on which it runs
	 * @param user
	 * @param passwordMD5Hex The {@link Codings.encodeMD5Hex}-encoded String conforming the following
	 * pattern: "user:password", where user normally is "admin" and password is the password
	 * for the specified user.
	 * @param cfg The global {@link Config} Object, which defines connection settings,
	 * verboseness of logging, etc.
	 * @throws <code>RemoteException</code> if a general yet unknown connection error occured
	 * @throws <code>PLoginWrongEx</code> if username / password don't match
	 * @throws <code>PSVNRevisionTooLowEx</code> if the SOAP-backend on the peer doesn't contain all referenced methods 
	 * @throws <code>PConnectionRefusedEx</code> if the host refused the connection, i.e. <code>port</code> is wrong or YaCy isn't running
	 * @throws <code>PServiceUnavailableEx</code> if YaCy was compiled without SOAP
	 * @throws <code>PConnectionResetEx</code> if the host resetted the connection
	 */
	public Peer(String host, int port, String user, String passwordMD5Hex, Config cfg)
			throws RemoteException, PLoginWrongEx, PSVNRevisionTooLowEx, PConnectionRefusedEx,
			PServiceUnavailableEx, PConnectionResetEx {
		this.cfg = cfg;
		this.useUserHeaderTable = cfg.useChunkedTransfer;
		connect(host, port, user, passwordMD5Hex);
	}
	
	private void fetchInitialValues() throws RemoteException, PLoginWrongEx, PSVNRevisionTooLowEx, 
			PConnectionRefusedEx, PServiceUnavailableEx, PConnectionResetEx {
		updateThread = new UpdateThread();
		updateThread.doSingleRun();						// initial load of values
		if (updateThread.getException() != null) {
			if (updateThread.getException().getMessage().contains(ERR_CONNECTION_REFUSED)) {
				throw new PConnectionRefusedEx();
			} else if (updateThread.getException().getMessage().contains(ERR_CONNECTION_RESET)) {
				throw new PConnectionResetEx();
			} else if (updateThread.getException().getMessage().indexOf(ERR_SERVICE_UNAVAILABLE) > -1 ||
					updateThread.getException().getMessage().indexOf(ERR_NOT_IMPLEMENTED) > -1) {
				throw new PServiceUnavailableEx();
			} else {
				throw updateThread.getException();
			}
		}
		this.name = this.getName();
		
		try {
			this.peerHash = this.getStatus().getPeerHash();
			this.svnRevision = Integer.valueOf(this.getAdmin().getConfigProperty("svnRevision")).intValue();
			this.blacklistTypes = this.getBlacklist().getBlacklistTypes();
			this.memoryMax = Long.valueOf(new DOMBuilder().build(this.getStatus().getStatus()).getRootElement().getChild("memory").getChildText("max"));
			Element d = new DOMBuilder().build(getAdmin().getTransferProperties()).getRootElement();
			this.indexDistributionEnabled = Boolean.parseBoolean(d.getChildText("allowDistributeIndex"));
			this.indexDistributionWhileCrawling = Boolean.parseBoolean(d.getChildText("allowDistributeIndexWhileCrawling"));
			this.indexReceiveEnabled = Boolean.parseBoolean(d.getChildText("allowIndexReceive"));
			this.indexReceiveBlockBlacklist = Boolean.parseBoolean(d.getChildText("indexReceiveBlockBlacklist"));
			String f = this.getAdmin().getConfigProperty("vString");
			this.version = f.substring(0, f.indexOf('/') - 1);
		} catch (RemoteException e) {
			if (e.getMessage().equals(ERR_LOGIN_REQUIRED))
				throw new PLoginWrongEx(pwdEnc);
			else
				cfg.log(e);
		}
		
		if (this.svnRevision < Config.neededSVNRevision) {
			cfg.log(String.format(WAR_SVN_TOO_LOW, this, this.svnRevision, Config.neededSVNRevision));
			if (cfg.noVersionCheck)
				cfg.log(INFO_SVN_TOO_LOW_OVERRIDDEN);
			else
				throw new PSVNRevisionTooLowEx(this.svnRevision, Config.neededSVNRevision);
		}
		this.connected = true;
		cfg.log.setNetworkIdle();
		firePeerEvent(new PeerConnectionEstablishedEvent());
	}
	
	/**
	 * The peer object initiates updates for the active plugin in a certain interval. With this
	 * method you can specify which plugin object shall be updated in which interval.
	 * 
	 * @see views.PanelPlugin
	 * @param active the active plugin
	 * @param newUpdate the update delay in milliseconds
	 */
	public void setUpdateProperties(PanelPlugin active, int newUpdate) /*throws RemoteException*/ {
		if ((updateThread == null || updateThread.isInterrupted()) && newUpdate > 0)
			updateThread = new UpdateThread();
		Object[] old = new Object[] { updateThread.getActivePlugin(), updateThread.getDelay() };
		updateThread.setActivePlugin(active);
		updateThread.setDelay(newUpdate);
		firePeerEvent(new PeerSettingChangedEvent(SC_UPDATE_PROPERTIES, old, new Object[] { active, newUpdate }));
		if (newUpdate > 0) {
			synchronized(updateThread) {
				if (updateThread.isAlive())
					updateThread.notify();
				else 
					updateThread.start();
			}
		}
	}
	
	public String toString() {
		return this.user + '@' + this.host + ':' + this.port;
	}
	
	/* -----------------------------------------------------------------------
	 * -------------------------- event functions ----------------------------
	 * ----------------------------------------------------------------------- */
	
	/**
	 * This method registeres a listener for all events this peer fires
	 * 
	 * @see yacy.event.PeerEventListener
	 * @param listener the PeerEventListener to be registered, which receives all
	 * {@link yacy.event.PeerEvent PeerEvents} afterwards
	 */
	public void registerPeerEventListener(PeerEventListener listener) {
		this.peerEventListeners.add(PeerEventListener.class, listener);
	}
	
	/**
	 * This method removes a listener. It will not receive events fired by this peer anymore.
	 * 
	 * @see yacy.event.PeerEventListener
	 * @param listener the PeerEventListener to be removed
	 */
	public void unregisterPeerEventListener(PeerEventListener listener) {
		this.peerEventListeners.remove(PeerEventListener.class, listener);
	}
	
	/**
	 * @see yacy.event.PeerEvent
	 * @param event the event to be fired or any from PeerEvent derived object; if the peer
	 * object, peer name or url of the event is not set, it will be set to the ones from this
	 * peer
	 */
	public void firePeerEvent(PeerEvent event) {
		if (event.getPeer() == null)
			event.setPeer(this);
		if (this.network != null) {
			if (event.getPeerName() == null)
				event.setPeerName(getName());
			if (event.getPeerURL() == null)
				event.setPeerURL(getHost() + ':' + getPort());
		}
		Object[] listeners = this.peerEventListeners.getListenerList();
		for (int i=0; i<listeners.length; i+=2)
			if (listeners[i] == PeerEventListener.class)
				((PeerEventListener)listeners[i + 1]).peerEvent(event);
	}
	
	public void firePeerInterruptableEvent(PeerInterruptableEvent event) throws PInterruptionEx {
		if (event.getPeer() == null)
			event.setPeer(this);
		if (this.network != null) {
			if (event.getPeerName() == null)
				event.setPeerName(getName());
			if (event.getPeerURL() == null)
				event.setPeerURL(getHost() + ':' + getPort());
		}
		Object[] listeners = this.peerEventListeners.getListenerList();
		for (int i=0; i<listeners.length; i+=2)
			if (listeners[i] == PeerEventListener.class) {
				((PeerEventListener)listeners[i + 1]).peerInterruptableEvent(event);
				if (event.isInterrupted())
					throw new PInterruptionEx(event);
			}
	}
	
	/* -----------------------------------------------------------------------
	 * ------------------------- private functions ---------------------------
	 * ----------------------------------------------------------------------- */
	
	private class UpdateThread extends Thread {
		
		private RemoteException e 		= null;
		private long 			lastrun	= 0;
		private int 			delay 	= 0;
		private PanelPlugin 	active 	= null;
		private boolean 		stopped = false;
		
		public PanelPlugin 		getActivePlugin() 	{ return this.active; }
		public int 				getDelay() 			{ return this.delay; }
		public RemoteException 	getException() 		{ return e; }
		public void 			resetException() 	{ e = null; }
		
		public void setActivePlugin(PanelPlugin active) { this.active = active; }
		public void setDelay(int delayms) 				{ this.delay = delayms; }
		
		public UpdateThread() {  }
		
		@SuppressWarnings("deprecation")
		public void doSingleRun() {
			if (this.lastrun != 0)
				cfg.logDebug("Last update was " + (System.currentTimeMillis() - this.lastrun) + " ms ago.");
			this.lastrun = System.currentTimeMillis();
			DOMBuilder b = new DOMBuilder();
			cfg.log.setNetworkBusy();
			try {
				network = b.build(getStatus().network());
				// properties = b.build(getAdmin().getConfigPropertyList());
				// transferProps = b.build(getAdmin().getTransferProperties());
				// queues = b.build(getStatus().getQueueStatus(queueCounts[INDEXER],queueCounts[LOADER],queueCounts[LOCAL],queueCounts[REMOTE]));
			} catch (RemoteException e) {
				this.e = e;
				// cfg.log("Error during connect", e);

				if (e.getMessage().indexOf(ERR_CONNECTION_REFUSED) > -1 ||
						e.getMessage().indexOf(ERR_CONNECTION_RESET) > -1 ||
						e.getMessage().indexOf(ERR_SERVICE_UNAVAILABLE) > -1) {
					// System.out.println("An Error occured, stopping update-thread");
					cfg.log.setNetworkError();
					this.interrupt();
				}
			} finally {
				cfg.log.setNetworkIdle();
			}
			if (this.active != null)
				this.active.doUpdate();
			firePeerEvent(new PluginUpdateEvent());
			if (cfg.logLevel == Config.LOG_DEBUG)
				System.out.println(" Finished update.");
		}
		
		public synchronized void run() {
			while (!stopped) {
				this.doSingleRun();
				if (this.isInterrupted()) {
					cfg.logDebug("Warning: update thread has been interrupted");
					return;
				}
				try {
					if (this.delay > 0) {
						this.wait(this.delay);
					} else {
						this.wait();
					}
				} catch (InterruptedException e) { /* ignore this */ }
			}
		}
	}
	
	private void setConnectionAddons(Stub service, boolean setAdminPwd) {
		if (setAdminPwd)
			service.setHeader(ANOMIC_HEADER, AUTHORIZATION, pwdEnc);
		if (!this.cfg.useHTTP11)
			service._setProperty(MessageContext.HTTP_TRANSPORT_VERSION, HTTPConstants.HEADER_PROTOCOL_V10);
		if (this.useUserHeaderTable)
			service._setProperty(HTTPConstants.REQUEST_HEADERS, userHeaderTable);
		if (this.cfg.useGZipEncoding) {
			service._setProperty(HTTPConstants.MC_GZIP_REQUEST, Boolean.TRUE);
			service._setProperty(HTTPConstants.MC_ACCEPT_GZIP, Boolean.TRUE);
		}
		service.setTimeout(cfg.timeout);
	}
	
	private AdminService getAdmin() {
		if (this.admin == null) {
			cfg.log.setNetworkBusy();
			try {
				AdminServiceServiceLocator adminLoc = new AdminServiceServiceLocator(); 
				adminLoc.setadminEndpointAddress(this.pHost + ADMIN_PATH);
				this.admin = adminLoc.getadmin();
				setConnectionAddons((Stub)this.admin, true);
			} catch (ServiceException e) {
				cfg.log(e);
			} finally {
				cfg.log.setNetworkIdle();
			}
		}
		return this.admin;
	}
	
	private BlacklistService getBlacklist() {
		if (this.blacklist == null) {
			cfg.log.setNetworkBusy();
			try {
				BlacklistServiceServiceLocator blacklistLoc = new BlacklistServiceServiceLocator();
				blacklistLoc.setblacklistEndpointAddress(this.pHost + BLACKLIST_PATH);
				this.blacklist = blacklistLoc.getblacklist();
				setConnectionAddons((Stub)this.blacklist, true);
			} catch (ServiceException e) {
				cfg.log(e);
			} finally {
				cfg.log.setNetworkIdle();
			}
		}
		return this.blacklist;
	}
	
	private BookmarkService getBookmarks() {
		if (this.bookmark == null) {
			try {
				cfg.log.setNetworkBusy();
				BookmarkServiceServiceLocator bookmarkLoc = new BookmarkServiceServiceLocator();
				bookmarkLoc.setbookmarksEndpointAddress(this.pHost + BOOKMARKS_PATH);
				this.bookmark = bookmarkLoc.getbookmarks();
				setConnectionAddons((Stub)this.bookmark, true);
			} catch (ServiceException e) {
				cfg.log(e);
			} finally {
				cfg.log.setNetworkIdle();
			}
		}
		return this.bookmark;
	}
	
	private CrawlService getCrawl() {
		if (this.crawl == null) {
			cfg.log.setNetworkBusy();
			try {
				CrawlServiceServiceLocator crawlLoc = new CrawlServiceServiceLocator();
				crawlLoc.setcrawlEndpointAddress(this.pHost + CRAWL_PATH);
				this.crawl = crawlLoc.getcrawl();				
				setConnectionAddons((Stub)this.crawl, true);
			} catch (ServiceException e) {
				cfg.log(e);
			} finally {
				cfg.log.setNetworkIdle();
			}
		}
		return this.crawl;
	}
	
	private MessageService getMessages() {
		if (this.message == null) {
			cfg.log.setNetworkBusy();
			try {
				MessageServiceServiceLocator messageLoc = new MessageServiceServiceLocator();
				messageLoc.setmessagesEndpointAddress(this.pHost + MESSAGES_PATH);
				this.message = messageLoc.getmessages();
				setConnectionAddons((Stub)this.message, true);
			} catch (ServiceException e) {
				cfg.log(e);
			} finally {
				cfg.log.setNetworkIdle();
			}
		}
		return this.message;
	}
	
	private SearchService getSearch() {
		if (this.search == null) {
			cfg.log.setNetworkBusy();
			try {
				SearchServiceServiceLocator searchLoc = new SearchServiceServiceLocator();
				searchLoc.setsearchEndpointAddress(this.pHost + SEARCH_PATH);
				this.search = searchLoc.getsearch();
				setConnectionAddons((Stub)this.search, false);
			} catch (ServiceException e) {
				cfg.log(e);
			} finally {
				cfg.log.setNetworkIdle();
			}
		}
		return this.search;
	}
	
	private ShareService getShare() {
		if (this.share == null) {
			cfg.log.setNetworkBusy();
			try {
				ShareServiceServiceLocator shareLoc = new ShareServiceServiceLocator();
				shareLoc.setshareEndpointAddress(this.pHost + SHARE_PATH);
				this.share = shareLoc.getshare(); 
				setConnectionAddons((Stub)this.share, true);
			} catch (ServiceException e) {
				cfg.log(e);
			} finally {
				cfg.log.setNetworkIdle();
			}
		}
		return this.share;
	}
	
	private StatusService getStatus() {
		if (this.status == null) {
			cfg.log.setNetworkBusy();
			try {
				StatusServiceServiceLocator statusLoc = new StatusServiceServiceLocator();
				statusLoc.setstatusEndpointAddress(this.pHost + STATUS_PATH);
				this.status = statusLoc.getstatus(); 
				setConnectionAddons((Stub)this.status, true);
			} catch (ServiceException e) {
				cfg.log(e);
			} finally {
				cfg.log.setNetworkIdle();
			}
		}
		return this.status;
	}
	
	/* -----------------------------------------------------------------------
	 * ------------------------- helper functions ----------------------------
	 * ----------------------------------------------------------------------- */
	
	public String getPasswordMD5Hex() {
		return this.pwdEnc;
	}
	
	@SuppressWarnings("unchecked")
	private List<Element> getBlacklistsList(boolean refresh) throws RemoteException {
		if (refresh || this.blist == null) {
			cfg.log.setNetworkBusy();
			try {
				this.blist = new DOMBuilder().build(this.getBlacklist().getBlacklistList()).getRootElement().getChildren();
			} finally {
				cfg.log.setNetworkIdle();
			}
		}
		return this.blist;
	}
	
	private void setTransferProperties() throws RemoteException {
		cfg.log.setNetworkBusy();
		try {
			getAdmin().setTransferProperties(
					this.indexDistributionEnabled,
					this.indexDistributionWhileCrawling,
					this.indexReceiveEnabled,
					this.indexReceiveBlockBlacklist);
		} finally {
			cfg.log.setNetworkIdle();
		}
	}
	
	/* -----------------------------------------------------------------------
	 * --------------------------- get functions -----------------------------
	 * ----------------------------------------------------------------------- */
	
	/**
	 * @param name specifies the blacklist name for which the enabled services shall be
	 * returned
	 * @see Peer#getBlacklistTypes() for a list of available services a blacklist may be
	 * enabled for
	 * @throws RemoteException if the requested information could not be received from the
	 * connected peer
	 * @return all attributes the blacklist is enabled for
	 */
	public ArrayList<String> getBlacklistAttributes(String name) throws RemoteException {
		ArrayList<String> r = new ArrayList<String>();
		cfg.log.setNetworkBusy();
		List<Element> l = null;
		try {
			 l = getBlacklistsList(false);
		} finally {
			cfg.log.setNetworkIdle();
		}
		for (Element d : l) {
			if (d.getAttribute("name").getValue().equals(name)) {
				for (int i=0; i<this.blacklistTypes.length; i++)
					try {
						if (d.getAttribute(this.blacklistTypes[i]).getIntValue() == 1)
							r.add(this.blacklistTypes[i]);
					} catch (DataConversionException e) {  }
				return r;
			}
		}
		return r;
	}
	
	/**
	 * YaCy can use other blacklist backends than it's own {@value Peer#BLACKLIST_ENGINE_YACY},
	 * which may have different syntax, so remember to check for the uses backend.
	 * 
	 * @return the classpath of the currently set blacklist engine
	 */
	public String getBlacklistEngine() { return this.blacklistEngine; }
	
	public ArrayList<String> getBlacklists() throws RemoteException {
		List<Element> l = null;
		cfg.log.setNetworkBusy();
		try {
			l = this.getBlacklistsList(false);
		} finally {
			cfg.log.setNetworkIdle();
		}
		ArrayList<String> r = new ArrayList<String>();
		for (Element d : l)
			r.add(d.getAttributeValue("name"));
		return r;
	}
	
	/** 
	 * @param refresh if set, the contents will be fetched from the peer, otherwise the previously
	 * fetched information will be returned. Since blacklists may become very large, it is not
	 * reasonable to get the contents frequently
	 * @return the contents of all blacklists by the name of the blacklist
	 * @throws RemoteException if the requested information could not be received from the
	 * connected peer
	 */
	@SuppressWarnings("unchecked")
	public HashMap<String,String[]> getBlacklistsContents(boolean refresh) throws RemoteException {
		HashMap<String,String[]> r = new HashMap<String,String[]>();
		List<Element> l = null;
		cfg.log.setNetworkBusy();
		try {
			l = getBlacklistsList(refresh);
		} finally {
			cfg.log.setNetworkIdle();
		}
		for (Element d1 : l) {
			String[] s = new String[d1.getChildren().size()];
			int x=0;
			for (Element d2 : (List<Element>)d1.getChildren())
				s[x++] = d2.getText();
			r.put(d1.getAttributeValue("name"), s);
		}
		return r;
	}
	
	/**
	 * YaCy blacklists can be enabled for the following services (named "attributes" here):
	 * <ul>
	 * <li>Crawler</li>
	 * <li>Proxy</li>
	 * <li>Search</li>
	 * <li>DHT</li>
	 * </ul>
	 * 
	 * @return string-representations of the available services a blacklist may be enabled for
	 */
	public String[] getBlacklistTypes() { return this.blacklistTypes; }

	@SuppressWarnings("unchecked")
	public HashMap<String,Integer> getBookmarkTags() throws RemoteException {
		HashMap<String,Integer> r = new HashMap<String,Integer>();
		List<Element> l = null;
		cfg.log.setNetworkBusy();
		try {
			l = (List<Element>)new DOMBuilder().build(getBookmarks().getBookmarkTagList()).getRootElement().getChildren();
		} finally {
			cfg.log.setNetworkIdle();
		}
		for (Element d : l)
			r.put(d.getAttributeValue("tag"), Integer.parseInt(d.getAttributeValue("count")));
		return r;
	}
	
	/**
	 * Bookmarks are categorized in so-called tags. This method fetches all bookmarks belonging
	 * to one tag
	 * 
	 * @param tag is the name of the desired tag. If set to <code>null</code>, bookmarks from
	 * all tags - so the complete list - will be returned.
	 * @return all bookmarks that have been tagged with the given tag-name
	 * @throws RemoteException if the requested information could not be received from the
	 * connected peer
	 */
	@SuppressWarnings("unchecked")
	public ArrayList<Bookmark> getBookmarksFrom(String tag) throws RemoteException {
		ArrayList<Bookmark> r = new ArrayList<Bookmark>();
		List<Element> l = null;
		cfg.log.setNetworkBusy();
		try {
			l = (List<Element>)new DOMBuilder().build(getBookmarks().getBookmarkList(tag, null)).getRootElement().getChildren();
		} finally {
			cfg.log.setNetworkIdle();
		}
		for (Element d : l)
			r.add(new Bookmark(
					d.getAttributeValue("description"),
					d.getAttributeValue("extended"),
					d.getAttributeValue("href"),
					d.getAttributeValue("hash"),
					d.getAttributeValue("time"),
					d.getAttributeValue("tag").split(" ")));
		return r;
	}
	
	/**
	 * @return unorderedly all bookmarks from this YaCy peer
	 * @throws RemoteException if the requested information could not be received from the
	 * connected peer
	 */
	public ArrayList<Bookmark> getBookmarksAll() throws RemoteException {
		return getBookmarksFrom(null);
	}
	
	/**
	 * @return the pause state of all three queues (local crawler -, remote triggered crawl -
	 * and global crawl trigger queue). If one or more are running, the function will return
	 * false
	 * @throws RemoteException if the requested information could not be received from the
	 * connected peer
	 */
	public boolean isCrawlingPaused() throws RemoteException {
		Element r = null; 
		cfg.log.setNetworkBusy();
		try {
			r = new DOMBuilder().build(getCrawl().getCrawlPauseResumeState()).getRootElement();
		} finally {
			cfg.log.setNetworkIdle();
		}
		return (Boolean.parseBoolean(r.getChildText("localCrawl")) &&
				Boolean.parseBoolean(r.getChildText("remotetriggeredcrawl")) &&
				Boolean.parseBoolean(r.getChildText("globalcrawltrigger")));
	}
	
	/**
	 * @param name the name of the property
	 * @return the value belonging to the given property
	 * @throws RemoteException if the requested information could not be received from the
	 * connected peer
	 */
	public String getConfigProperty(String name) throws RemoteException {
		String r = null;
		cfg.log.setNetworkBusy();
		try {
			r = this.getAdmin().getConfigProperty(name);
		} finally {
			cfg.log.setNetworkIdle();
		}
		return r;
	}
	
	/*
	public int getPPMofActives() {
	}
	
	public int getMaxURLsInRam() {
	}
	
	public int getDHTCacheSize() {
	}*/
	
	public String getHash() {
		return this.peerHash;
	}
	
	/**
	 * @return the hostname the connected yacy peer runs on
	 */
	public String getHost() { return this.host; }
	
	/**
	 * @return the amount of URLs this peers knows about
	 */
	public long getLinks() {
		String dnr = this.network.getRootElement().getChild("your").getChildText("links");
		if (dnr.equals("-"))
			return 0L;
		return Long.parseLong(dnr.replaceAll("\\.", ""));
	}
	
	/**
	 * This method fetches the amount of known URLs by peers in parts of the global YaCy
	 * network.
	 * 
	 * @see Peer#PEER_STATI
	 * @param peerStatus specifies the part of the network. Must be a constant of {@link Peer#PEER_STATI}.
	 * @return the amount of known URLs in the specified part of the YaCy network
	 */
	public long getLinks(int peerStatus) {
		if (peerStatus < 0 || peerStatus > Peer.PEER_STATI.length - 1)
			throw new IllegalArgumentException("peerStatus is not of Peer.peerStatus");
		String dnr = this.network.getRootElement().getChild(Peer.PEER_STATI[peerStatus]).getChildText("links");
		if (dnr.equals("-"))
			return 0L;
		return Long.parseLong(dnr.replaceAll("\\.", ""));
	}
	
	/**
	 * Java does not access the OS-specific memory allocation techniques but the user must
	 * specify how much RAM the process might occupy maximal. This value is set on start-up
	 * via command-line parameter.<br>
	 * This value doesn't change once the connection is established successfully.
	 * 
	 * @return the maximal RAM size YaCy may use on the host in MB as specified on start-up
	 * @throws RemoteException if the requested information could not be received from the
	 * connected peer
	 */
	public int getMemoryMaxMB() throws RemoteException {
		cfg.log.setNetworkBusy();
		String drn = null;
		try {
			drn = getAdmin().getConfigProperty("javastart_Xmx");
		} finally {
			cfg.log.setNetworkIdle();
		}
		return Integer.valueOf(drn.substring(3, drn.length()- 1)).intValue();
	}
	
	/**
	 * This value doesn't change once the connection is established successfully.
	 * 
	 * @return the maximal RAM size YaCy may use on the host in Byte as the software retrieves
	 * itself.
	 */
	public long getMemoryMax() {
		return this.memoryMax;
	}
	
	/**
	 * This value doesn't change once the connection is established successfully.
	 * 
	 * @return the initial RAM size YaCy allocated on start-up in Bytes
	 * @throws RemoteException if the requested information could not be received from the
	 * connected peer
	 */
	public long getMemoryTotal() throws RemoteException {
		String drn = null;
		cfg.log.setNetworkBusy();
		try {
			drn = new DOMBuilder().build(this.getStatus().getStatus()).getRootElement().getChild("memory").getChildText("total");
		} finally {
			cfg.log.setNetworkIdle();
		}
		return Long.valueOf(drn);
	}
	
	/**
	 * @return the size of free RAM in Bytes at the time this request is made
	 * @throws RemoteException if the requested information could not be received from the
	 * connected peer
	 */
	public long getMemoryFree() throws RemoteException {
		String drn = null;
		cfg.log.setNetworkBusy();
		try {
			drn = new DOMBuilder().build(this.getStatus().getStatus()).getRootElement().getChild("memory").getChildText("free");
		} finally {
			cfg.log.setNetworkIdle();
		}
		return Long.valueOf(drn);
	}
	
	/**
	 * YaCy offers the possibility to send messages to other YaCy peers and therefore to
	 * receive them as well.
	 * 
	 * @see Message
	 * @param messageID the unique number identifying the message on the peer
	 * @return the complete message belonging to the passed message-ID
	 * @throws RemoteException if the requested information could not be received from the
	 * connected peer
	 */
	public Message getMessage(String messageID) throws RemoteException {
		Element d = null;
		cfg.log.setNetworkBusy();
		try {
			d = new DOMBuilder().build(getMessages().getMessage(messageID)).getRootElement();
		} finally {
			cfg.log.setNetworkIdle();
		}
		return new Message(
				d.getChildText("subject"),
				d.getChildText("message"),
				d.getChild("from").getAttributeValue("hash"),
				d.getChildText("from"),
				d.getChildText("to"),
				d.getChildText("date"));
	}
	
	/**
	 * YaCy offers the possibility to send messages to other YaCy peers and therefore to
	 * receive them as well.
	 * 
	 * @see Message
	 * @return all received messages as an ArrayList&lt;Message&gt;
	 * @throws RemoteException if the requested information could not be received from the
	 * connected peer
	 */
	public ArrayList<Message> getMessagesAll() throws RemoteException {
		ArrayList<Message> r = new ArrayList<Message>();
		String[] l = null;
		cfg.log.setNetworkBusy();
		try {
			l = getMessages().getMessageIDs();
		} finally {
			cfg.log.setNetworkIdle();
		}
		for (String d : l)
			r.add(getMessage(d));
		return r;
	}
	
	/**
	 * Each YaCy peer has an individual name which serves as a .yacy-domain at the same time
	 * if you surf over the YaCy proxy.
	 * 
	 * @return the name of the connected peer.
	 */
	public String getName() {
		if (this.name != null) return this.name;
		this.name = this.network.getRootElement().getChild("your").getChildText("name");
		return this.name;
	}
	
	/**
	 * This method fetches the amount of known peers in parts of the global YaCy network.
	 * 
	 * @see Peer#PEER_STATI
	 * @param peerStatus specifies the part of the network. Must be a constant of {@link Peer#PEER_STATI}.
	 * @return the amount of known peers in the specified part of the YaCy network
	 */
	public int getPeerCount(int peerStatus) {
		if (peerStatus < 0 || peerStatus > Peer.PEER_STATI.length - 1)
			throw new IllegalArgumentException("peerStatus is not of Peer.peerStatus");
		String dnr = this.network.getRootElement().getChild(Peer.PEER_STATI[peerStatus]).getChildText("count");
		return Integer.valueOf(dnr).intValue();
	}
	
	public int getPEER_STATI() {
		return Integer.valueOf(this.network.getRootElement().getChildText("status")).intValue();
	}
	
	/**
	 * @return the port this YaCy peer listens for incoming connection on 
	 */
	public int getPort() { return this.port; }
	
	/**
	 * PPM stands for indexed Pages Per Minute and is an indicator for the speed of the peer
	 * including hardware configuration as well as database performance.
	 * 
	 * @return the PPM at the time the request is made
	 */
	public int getPPM() {
		String dnr = this.network.getRootElement().getChild("your").getChildText("ppm");
		return Integer.valueOf(dnr).intValue();
	}
	
	/**
	 * YaCy works with several queues for processing webpages internally of which of four
	 * the entries and other information may be retrieved.
	 * 
	 * @see Peer#QUEUE_LOCAL
	 * @see Peer#QUEUE_LOADER
	 * @see Peer#QUEUE_INDEXER
	 * @see Peer#QUEUE_REMOTE
	 * @see QueueEntry
	 * @param queueType the type of queue 
	 * @param count the number of entries to be returned maximal 
	 * @return the queue including entries as a {@link QueueEntry}-object
	 * @throws RemoteException if the requested information could not be received from the
	 * connected peer
	 */
	@SuppressWarnings("unchecked")
	public ArrayList<QueueEntry> getQueue(int queueType, int count) throws RemoteException {
		Element d = null;
		cfg.log.setNetworkBusy();
		try {
			d = new DOMBuilder().build(this.getStatus().getQueueStatus(0, 0, count, 0))
			.getRootElement().getChild(QUEUE_TYPES[queueType]);
		} finally {
			cfg.log.setNetworkIdle();
		}
		ArrayList<QueueEntry> r = new ArrayList<QueueEntry>();
		for (Element f : (List<Element>)d.getChildren("entry")) {
			r.add(new QueueEntry(
					f.getChildText("initiator"),
					Integer.parseInt(f.getChildText("depth")),
					f.getChildText("modified"),
					f.getChildText("anchor"),
					f.getChildText("url"),
					(queueType == QUEUE_INDEXER) ? (Long.parseLong(f.getChildText("size"))) : (0L),
					f.getChildText("hash"),
					Boolean.parseBoolean(f.getChildText("inProcess"))));
		}
		return r;
	}
	
	public HashMap<String,ArrayList<QueueEntry>> getQueues(int count) throws RemoteException {
		HashMap<String,ArrayList<QueueEntry>> r = new HashMap<String,ArrayList<QueueEntry>>();
		for (int i=0; i<QUEUE_TYPES.length; i++)
			r.put(QUEUE_TYPES[i], getQueue(i, count));
		return r;
	}
	
	/**
	 * The RAM Cache is the cache where all parsed words are stored as long as possible.
	 * A high value accellerates indexing. Attention: This function calls the peer.
	 * 
	 * @return the maximum count of words that can be held in RAM Cache. 
	 */
	public int getRamCacheSize() throws RemoteException {
		String drn = null;
		cfg.log.setNetworkBusy();
		try {
			drn = this.getAdmin().getConfigProperty("wordCacheMaxCount");
		} finally {
			cfg.log.setNetworkIdle();
		}
		return Integer.valueOf(drn);
	}
	
	/**
	 * @return the current size of the RAM cache, also known as DHT-out cache
	 * @throws RemoteException if the requested information could not be received from the
	 * connected peer
	 */
	public int getRamCacheNow() throws RemoteException {
		String drn = null;
		cfg.log.setNetworkBusy();
		try {
			drn = new DOMBuilder().build(this.getStatus().getStatus()).getRootElement().getChildText("wordCacheSize");
		} finally {
			cfg.log.setNetworkIdle();
		}
		return Integer.valueOf(drn);
	}
	
	/**
	 * @return the version-part of the version/SVN-combination (may not be a value, i.e. "dev")
	 */
	public String getReleaseVersion() { return this.version; }
	
	/**
	 * Retrieves a directory-listing from the peer
	 * 
	 * @param folder The folder whose content shall be fetched 
	 * @return A {@link java.util.ArrayList ArrayList} containing all elements the
	 * specified share-folder contains 
	 * @throws RemoteException if the requested information could not be received from the
	 * connected peer
	 */
	@SuppressWarnings("unchecked")
	public ArrayList<ShareElement> getShareDirList(String folder) throws RemoteException {
		ArrayList<ShareElement> r = new ArrayList<ShareElement>();
		List<Element> l = null;
		cfg.log.setNetworkBusy();
		try {
			l = (List<Element>)new DOMBuilder().build(this.getShare().getDirList(folder)).getRootElement().getChildren();
		} finally {
			cfg.log.setNetworkIdle();
		}
		for (Element d : l) {
			try {
				if (d.getName().equals("directory"))
					r.add(new ShareFolder(
							d.getChildText("name"),
							d.getChild("date").getAttribute("timestamp").getLongValue()));
				else if (d.getName().equals("file"))
					r.add(new ShareFile(
							d.getChildText("name"),
							d.getChild("date").getAttribute("timestamp").getLongValue(),
							d.getChild("size").getAttribute("bytes").getLongValue(),
							d.getChildText("comment"),
							d.getChildText("yacyURL")));
			} catch (DataConversionException e) {  }
		}
		return r;
	}
	
	/**
	 * @return the SVN revision number of the version of this peer
	 */
	public int getSVNVersion() { return this.svnRevision; }
	
	@SuppressWarnings("unchecked")
	public ArrayList<TransferAction> getTransferActionsTillNow() throws RemoteException {
		ArrayList<Element> logEntries = new ArrayList<Element>();
		org.w3c.dom.Document dom = null;
		long seq;
		cfg.log.setNetworkBusy();
		try {
			dom = getAdmin().getServerLog(fromLogSequence);
		} finally {
			cfg.log.setNetworkIdle();
		}
		if (dom != null) {
			for (Element el : (List<Element>)new DOMBuilder().build(dom).getRootElement().getChildren("record")) {
				seq = Long.valueOf(el.getChildText("sequence"));
				if (fromLogSequence < seq)
					fromLogSequence = seq;
				logEntries.add(el);
			}
		}
		return TransferAction.genTransferActions(logEntries);
	}
	
	/**
	 * There are four different connection states a peer may have depending on router
	 * configuration, network connection and/or seed-upload settings.
	 * 
	 * @see Peer#STATE_VIRGIN
	 * @see Peer#STATE_JUNIOR
	 * @see Peer#STATE_SENIOR
	 * @see Peer#STATE_PRINCIPAL
	 * @return the current state of the YaCy peer in the network
	 */
	public String getType() {
		return network.getRootElement().getChild("your").getChildText("type");
	}
	
	/**
	 * @return how long the peer has been running in the following format:
	 * "<code>$x d $x h $x m</code>" where <code>$x</code> stands for a valid value
	 * for the particular field. If the uptime is shorter than 24 hours, the day-part
	 * won't be part of the uptime
	 */
	public String getUptime() {
		return this.network.getRootElement().getChild("your").getChildText("uptime");
	}
	
	/**
	 * @see yacy.datatypes.UrlInfo#HTML_ESCAPED
	 * @see yacy.datatypes.UrlInfo#HTML_NORMAL
	 * @see yacy.datatypes.UrlInfo#SINGLE_SENTENCES
	 * @param url
	 * @param mode
	 * @return
	 * @throws RemoteException if the requested information could not be received from the
	 * connected peer
	 */
	@SuppressWarnings("unchecked")
	public UrlInfo getUrlInfo(String url, int mode) throws RemoteException {
		Element body = null;
		cfg.log.setNetworkBusy();
		try {
			body = new DOMBuilder().build(this.getSearch().urlInfo(url, mode)).getRootElement();
		} finally {
			cfg.log.setNetworkIdle();
		}
		if (body == null) return null;
		Element meta = body.getChild("meta");
		body = body.getChild("body");
		switch (mode) {
			case UrlInfo.HTML_ESCAPED:
			case UrlInfo.HTML_NORMAL:
				return new UrlInfo(
						meta.getChildText("url"),
						meta.getChildText("hash"),
						Integer.valueOf(meta.getChildText("wordcount")),
						meta.getChildText("desc"),
						Long.valueOf(meta.getChildText("size")),
						meta.getChildText("mimeType"),
						Integer.valueOf(meta.getChildText("statusCode")),
						meta.getChildText("statusText"),
						Integer.valueOf(body.getAttributeValue("mode")),
						body.getText());
				
			case UrlInfo.SINGLE_SENTENCES:
				ArrayList<String> r = new ArrayList<String>();
				for (Element d : (List<Element>)body.getChildren())
					r.add(d.getText());
				return new UrlInfo(
						meta.getChildText("url"),
						meta.getChildText("hash"),
						Integer.valueOf(meta.getChildText("wordcount")),
						meta.getChildText("desc"),
						Long.valueOf(meta.getChildText("size")),
						meta.getChildText("mimeType"),
						Integer.valueOf(meta.getChildText("statusCode")),
						meta.getChildText("statusText"),
						Integer.valueOf(body.getAttributeValue("mode")),
						r);
				
			default:
				return null;
		}
	}
	
	/**
	 * @see Peer#getUrlInfo(String, int)
	 * @see yacy.datatypes.UrlInfo#SINGLE_SENTENCES
	 * @param url
	 * @return
	 * @throws RemoteException if the requested information could not be received from the
	 * connected peer
	 */
	public UrlInfo getUrlInfo(String url) throws RemoteException {
		return getUrlInfo(url, UrlInfo.SINGLE_SENTENCES);
	}
	
	public String getUser() {
		return this.user;
	}
	
	/**
	 * @return how many different RWI-entities this peer knows
	 */
	public long getWords() {
		String dnr = this.network.getRootElement().getChild("your").getChildText("words");
		if (dnr.equals("-"))
			return 0L;
		return Long.parseLong(dnr.replaceAll("\\.", ""));
	}
	
	/**
	 * This method fetches the amount of known RWIs by peers in parts of the global YaCy
	 * network.
	 * 
	 * @see Peer#PEER_STATI
	 * @param peerStatus specifies the part of the network. Must be a constant of {@link Peer#PEER_STATI}.
	 * @return the amount of known RWIs in the specified part of the YaCy network
	 */
	public long getWords(int peerStatus) {
		if (peerStatus < 0 || peerStatus > Peer.PEER_STATI.length - 1)
			throw new IllegalArgumentException("peerStatus is not of Peer.peerStatus");
		String dnr = this.network.getRootElement().getChild(Peer.PEER_STATI[peerStatus]).getChildText("words");
		if (dnr.equals("-"))
			return 0L;
		return Long.parseLong(dnr.replaceAll("\\.", ""));
	}
	
	/* -----------------------------------------------------------------------
	 * -------------------------- other functions ----------------------------
	 * ----------------------------------------------------------------------- */
	
	/**
	 * This method adds an entry to the blacklist specified the <code>file</code>-
	 * argument. No check is done to prevent "wrong" entries which are unusable for
	 * the current blacklist-backend or double entries to be entered.
	 * 
	 * @param entry an URL-representation possibly containing wildcards (<code>*</code>)
	 * formatted according the selected blacklist-engine
	 * @param file the blacklist-file where entry shall be added
	 * @throws RemoteException if the requested operation could not be performed on the
	 * connected peer
	 */
	public void addToBlacklist(String entry, String file) throws RemoteException {
		cfg.log.setNetworkBusy();
		try {
			this.getBlacklist().addBlacklistItem(file, entry);
			firePeerEvent(new PeerSettingChangedEvent(
					String.format(SC_ADDED_BLACKLIST_ENTRY_TO, file),
					null,
					entry));
		} finally {
			cfg.log.setNetworkIdle();
		}
	}
	
	/**
	 * @see Peer#getBlacklistTypes()
	 * @param name the name of the new blacklist
	 * @param shared specify whether the blacklist shall be accessible from other peers
	 * @param activateForBLTypes contains all service names the blacklist shall be
	 * enabled for
	 * @throws IllegalArgumentException if activateForBLTypes contains illegal service-
	 * names which are not supported by the connected peer
	 * @throws PBlacklistExistsEx if a blacklist with the specified name
	 * already exists on the peer and therefore cannot be created
	 * @throws RemoteException if the requested operation could not be performed on the
	 * connected peer
	 */
	public void addBlacklist(String name, boolean shared, String[] activateForBLTypes)
			throws PBlacklistExistsEx, RemoteException {
		boolean found;
		for (String d : activateForBLTypes) {
			found = false;
			for (String c : this.blacklistTypes)
				if (d.equals(c)) {
					found = true;
					break;
				}
			if (!found)
				throw new IllegalArgumentException("activateForBLTypes contains illegal blacklist type: " + d);
		}
		try {
			cfg.log.setNetworkBusy();
			this.getBlacklist().createBlacklist(name, shared, activateForBLTypes);
			firePeerEvent(new PeerSettingChangedEvent("Added blacklist", null, name));
		} catch (RemoteException e) {
			if (e instanceof AxisFault &&
					((AxisFault)e).getFaultString().equals("Blacklist with name '" + name + "' already exist."))
				throw new PBlacklistExistsEx(name, this);
			else
				throw e;
		} finally {
			cfg.log.setNetworkIdle();
		}
	}
	
	/**
	 * @see Bookmark
	 * @param title the title of the new bookmark
	 * @param description the description what the bookmarked site is about
	 * @param url the URL to the site
	 * @param tags keywords to categorize bookmarks
	 * @param isPublic specify whether this bookmark shall be visible from other peers
	 * @return the new Bookmark item
	 * @throws RemoteException if the requested operation could not be performed on the
	 * connected peer
	 */
	public Bookmark addBookmark(String title, String description, String url,
			String[] tags, boolean isPublic) throws RemoteException {
		String s = null;
		Bookmark r = null;
		cfg.log.setNetworkBusy();
		try {
		     s = getBookmarks().addBookmark(url, title, description, tags, isPublic);
		     r = new Bookmark(title, description, url, s, Long.toString(System.currentTimeMillis()), tags);
		     firePeerEvent(new PeerSettingChangedEvent(SC_ADDED_BOOKMARK, null, r));
		} finally {
			cfg.log.setNetworkIdle();
		}
		return r;
	}
	
	/**
	 * @see Bookmark
	 * @param bookmark the Bookmark item to add
	 * @param isPublic specify whether this bookmark shall be visible from other peers
	 * @return the new Bookmark item, which might be different from the passed one,
	 * because it definitely contains an {@link Bookmark#getUrlHash() URL-Hash} whereas
	 * the passed doesn't need to contain one
	 * @throws RemoteException if the requested operation could not be performed on the
	 * connected peer
	 */
	public Bookmark addBookmark(Bookmark bookmark, boolean isPublic) throws RemoteException {
		String s = null;
		Bookmark r = null;
		cfg.log.setNetworkBusy();
		try {
			s = getBookmarks().addBookmark(
					bookmark.getURL(),
					bookmark.getTitle(),
					bookmark.getDescription(),
					bookmark.getTags(),
					isPublic);
			r = new Bookmark(
					bookmark.getTitle(),
					bookmark.getDescription(),
					bookmark.getURL(),
					s,
					Long.toString(System.currentTimeMillis()),
					bookmark.getTags());
			firePeerEvent(new PeerSettingChangedEvent(SC_ADDED_BOOKMARK, null, r));
		} finally {
			cfg.log.setNetworkIdle();
		}
		return r;
	}
	
	
	/*
	// TODO !
	public void crawl(
			String crawlingURL,
			String crawljobTitel,
			int crawlingDepth,
			String crawlingFilter,
			boolean localIndexing,
			boolean crawlingQ,
			boolean storeHTCache,
			boolean crawlOrder,
			String crawlOrderIntention,
			boolean xsstopw) throws RemoteException {
		Element d = new DOMBuilder().build(getCrawl().crawling(
				crawlingURL,
				crawljobTitel,
				crawlingDepth,
				crawlingFilter,
				localIndexing,
				crawlingQ,
				storeHTCache,
				crawlOrder,
				crawlOrderIntention,
				xsstopw)).getRootElement();
	}*/
	
	public void createShareDir(String pathToDir) throws RemoteException {
		String dir = "/";
		String[] toks = ((pathToDir.charAt(0) == '/') ? (pathToDir) : ("/" + pathToDir)).split("/");
		cfg.log.setNetworkBusy();
		try {
			for (int i=1; i<toks.length; dir += toks[i++] + "/") {
				getShare().createDirectory(dir, toks[i]);
			}
			firePeerEvent(new PeerSettingChangedEvent(
					SC_CREATED_SHARE_DIR,
					null,
					pathToDir));
		} finally {
			cfg.log.setNetworkIdle();
		}
	}
	
	/**
	 * @param url the URL of the bookmark to remove
	 * @throws RemoteException if the requested operation could not be performed on the
	 * connected peer
	 */
	public void deleteBookmark(String url) throws RemoteException {
		cfg.log.setNetworkBusy();
		try {
			getBookmarks().deleteBookmark(url);
			firePeerEvent(new PeerSettingChangedEvent(
					SC_DELETED_BOOKMARK,
					url,
					null));
		} finally {
			cfg.log.setNetworkIdle();
		}
	}
	
	/**
	 * @param bookmark the bookmark to remove
	 * @throws RemoteException if the requested operation could not be performed on the
	 * connected peer
	 */
	public void deleteBookmark(Bookmark bookmark) throws RemoteException {
		cfg.log.setNetworkBusy();
		try {
			getBookmarks().deleteBookmark(bookmark.getURL());
			firePeerEvent(new PeerSettingChangedEvent(
					SC_DELETED_BOOKMARK,
					bookmark,
					null));
		} finally {
			cfg.log.setNetworkIdle();
		}
	}
	
	/** 
	 * @param urls the URLs of the bookmarks to remove
	 * @throws RemoteException if the requested operation could not be performed on the
	 * connected peer
	 */
	public void deleteBookmarks(String[] urls) throws RemoteException {
		cfg.log.setNetworkBusy();
		try {
			getBookmarks().deleteBookmarks(urls);
			firePeerEvent(new PeerSettingChangedEvent(
					SC_DELETED_BOOKMARKS,
					urls,
					null));
		} finally {
			cfg.log.setNetworkIdle();
		}
	}
	
	/**
	 * @param bookmarks the bookmarks to remove
	 * @throws RemoteException if the requested operation could not be performed on the
	 * connected peer
	 */
	public void deleteBookmarks(Bookmark[] bookmarks) throws RemoteException {
		ArrayList<String> c = new ArrayList<String>();
		for (Bookmark d : bookmarks)
			c.add(d.getURL());
		cfg.log.setNetworkBusy();
		try {
			getBookmarks().deleteBookmarks(c.toArray(new String[c.size()]));
			firePeerEvent(new PeerSettingChangedEvent(
					SC_DELETED_BOOKMARKS,
					bookmarks,
					null));
		} finally {
			cfg.log.setNetworkIdle();
		}
	}
	
	/**
	 * @param messageID the message ID of the message to remove
	 * @throws RemoteException if the requested operation could not be performed on the
	 * connected peer
	 */
	public void deleteMessage(String messageID) throws RemoteException {
		cfg.log.setNetworkBusy();
		try {
			getMessages().deleteMessage(messageID);
			firePeerEvent(new PeerSettingChangedEvent(SC_DELETED_MESSAGE, messageID, null));
		} finally {
			cfg.log.setNetworkIdle();
		}
	}
	
	/**
	 * @param messageIDs the message IDs of the messages to remove
	 * @throws RemoteException if the requested operation could not be performed on the
	 * connected peer
	 */
	public void deleteMessages(String[] messageIDs) throws RemoteException {
		cfg.log.setNetworkBusy();
		try {
			getMessages().deleteMessages(messageIDs);
			firePeerEvent(new PeerSettingChangedEvent(SC_DELETED_MESSAGES, messageIDs, null));
		} finally {
			cfg.log.setNetworkIdle();
		}
	}
	
	/**
	 * @param pathToFile absolute path to the file on remote side to remove
	 * @throws RemoteException if the requested operation could not be performed on the
	 * connected peer
	 */
	public void deleteShareFile(String pathToFile) throws RemoteException {
		cfg.log.setNetworkBusy();
		try {
			getShare().delete(
					pathToFile.substring(0, pathToFile.lastIndexOf('/') + 1),
					pathToFile.substring(pathToFile.lastIndexOf('/') + 1));
			firePeerEvent(new PeerSettingChangedEvent(SC_DELETED_SHARE_FILE, pathToFile, null));
		} finally {
			cfg.log.setNetworkIdle();
		}
	}
	
	/**
	 * @param bookmark the bookmark to edit
	 * @param newTitle the new title for this bookmark, if set to <code>null</code>, the
	 * one from bookmark is being used
	 * @param newDescription the new description for this bookmark, if set to <code>null</code>,
	 * the one from bookmark is being used
	 * @param newURL the new URL for this bookmark, if set to <code>null</code>, the
	 * one from bookmark is being used
	 * @param newTags the new tags for this bookmark, if set to <code>null</code>, the
	 * one from bookmark is being used
	 * @param isPublic specify whether the bookmark shall be accessible by other peers
	 * @return the edited bookmark
	 * @throws IllegalArgumentException if the bookmark is not valid (i.e. contains no URL Hash)
	 * @throws RemoteException if the requested operation could not be performed on the
	 * connected peer
	 */
	public Bookmark editBookmark(Bookmark bookmark, String newTitle, String newDescription,
			String newURL, String[] newTags, boolean isPublic) throws RemoteException {
		String s = null;
		Bookmark r = null;
		if (bookmark == null || bookmark.getUrlHash() == null)
			throw new IllegalArgumentException("bookmark is not a valid bookmark, urlHash is missing.");
		
		if (newTitle == null) newTitle = bookmark.getTitle();
		if (newDescription == null) newDescription = bookmark.getDescription();
		if (newURL == null) newURL = bookmark.getURL();
		if (newTags == null) newTags = bookmark.getTags();
		cfg.log.setNetworkBusy();
		try  {
			s = getBookmarks().editBookmark(bookmark.getUrlHash(), newURL, newTitle, newDescription, newTags, isPublic);
			r = new Bookmark(newTitle, newDescription, newURL, s, bookmark.getDate(), newTags);
			firePeerEvent(new PeerSettingChangedEvent(SC_EDITED_BOOKMARK, bookmark, r));
		} finally {
			cfg.log.setNetworkIdle();
		}
			
		return r;
	}
	
	/**
	 * @param pathToFile the absolute path to the file on remote side
	 * @param newComment the new comment for the file
	 * @param indexFile specify whether the comment should be indexed or not
	 * @throws RemoteException if the requested operation could not be performed on the
	 * connected peer
	 */
	public void editShareFileComment(
			String pathToFile,
			String newComment,
			boolean indexFile) throws RemoteException {
		cfg.log.setNetworkBusy();
		try {
			getShare().changeComment(
					pathToFile.substring(0, pathToFile.lastIndexOf('/') + 1),
					pathToFile.substring(pathToFile.lastIndexOf('/') + 1),
					newComment,
					indexFile);
			firePeerEvent(new PeerSettingChangedEvent(
					String.format(SC_EDITED_SHARE_FILE_COMMENT, pathToFile),
					null,
					newComment));
		} finally {
			cfg.log.setNetworkIdle();
		}
	}
	
	/**
	 * Shared blacklists are accesible for other peers so that they can enhance their own
	 * blacklist by adding entries of this one.
	 * 
	 * @param name the name of the blacklist
	 * @return whether the specified blacklist is shared with all other yacy peers
	 * @throws RemoteException if the requested operation could not be performed on the
	 * connected peer
	 */
	public boolean isBlacklistShared(String name) throws RemoteException {
		for (Element d : getBlacklistsList(false)) {
			if (d.getAttribute("name").getValue().equals(name))
				try {
					return (d.getAttribute("shared").getIntValue() == 1);
				} catch (DataConversionException e) { return false; }
		}
		return false;
	}
	
	public boolean isConnected() {
		return this.connected;
	}
	
	/**
	 * By distributing the local index to other YaCy peers, the efficency for searches increases.
	 * It can be disabled i.e. to give full speed to the crawler/indexer but should generally
	 * stay enabled, otherwise huge parts of the index will not be accessed because it resides
	 * at the wrong peer. 
	 * 
	 * @return whether the DHT-distribution is enabled
	 * @throws RemoteException if the requested information could not be received from the
	 * connected peer
	 */
	public boolean isIndexDistributionEnabled() throws RemoteException {
		cfg.log.setNetworkBusy();
		try {
			this.indexDistributionEnabled = Boolean.parseBoolean(
					new DOMBuilder().build(getAdmin().getTransferProperties())
					.getRootElement().getChildText("allowDistributeIndex"));
		} finally {
			cfg.log.setNetworkIdle();
		}
		return this.indexDistributionEnabled;
	}
	
	/**
	 * By default the index distribution is not enabled during a crawl-job because it generates
	 * high database load. This option provides the possibility to enable it nevertheless. 
	 * 
	 * @return whether DHT-distribution is enabled during crawls
	 * @throws RemoteException if the requested information could not be received from the
	 * connected peer
	 */
	public boolean isIndexDistributionEnabledWhileCrawling() throws RemoteException {
		cfg.log.setNetworkBusy();
		try {
			this.indexDistributionWhileCrawling = Boolean.parseBoolean(
					new DOMBuilder().build(getAdmin().getTransferProperties()).getRootElement()
					.getChildText("allowDistributeIndexWhileCrawling"));
		} finally {
			cfg.log.setNetworkIdle();
		}
		return this.indexDistributionWhileCrawling;
	}
	
	/**
	 * The counter part of index distribution is index receive. This option should generally
	 * be enabled, otherwise you will on the one hand not be able to search using remote peers
	 * and on the other hand you cannot help the YaCy network this way.<br>
	 * Other peers will transfer those parts of their indexes to this peer, which fit into the
	 * global DHT-scheme.
	 * 
	 * @return whether index receive is enabled
	 * @throws RemoteException if the requested information could not be received from the
	 * connected peer
	 */
	public boolean isIndexReceiveEnabled() throws RemoteException {
		cfg.log.setNetworkBusy();
		try {
			this.indexReceiveEnabled = Boolean.parseBoolean(
					new DOMBuilder().build(getAdmin().getTransferProperties()).getRootElement()
					.getChildText("allowReceiveIndex"));
		} finally {
			cfg.log.setNetworkIdle();
		}
		return this.indexReceiveEnabled;
	}
	
	/**
	 * Though index receive is very important for the YaCy network, it might be preferrable to
	 * filter out sites which match the local blacklists enabled for DHT to be sure that the
	 * index only contains those URLs which are not blocked. By default this option is set to
	 * false.
	 * 
	 * @return whether a blacklist check should be performed on the received DHT chunks
	 * @throws RemoteException if the requested information could not be received from the
	 * connected peer
	 */
	public boolean isIndexReceiveBlockBlacklistEnabled() throws RemoteException {
		cfg.log.setNetworkBusy();
		try {
			this.indexReceiveBlockBlacklist = Boolean.parseBoolean(
					new DOMBuilder().build(getAdmin().getTransferProperties()).getRootElement()
					.getChildText("indexReceiveBlockBlacklist"));
		} finally {
			cfg.log.setNetworkIdle();
		}
		return this.indexReceiveBlockBlacklist;
	}
	
	/**
	 * @param item the entry-String to remove; must exactly match the entry on remote side,
	 * including wildcards
	 * @param blacklistName the name of the blacklist where this entry shall be removed
	 * @throws RemoteException if the requested operation could not be performed on the
	 * connected peer
	 */
	public void removeBlacklistItem(String item, String blacklistName) throws RemoteException {
		cfg.log.setNetworkBusy();
		try {
			this.getBlacklist().removeBlacklistItem(blacklistName, item);
			firePeerEvent(new PeerSettingChangedEvent(String.format(SC_REMOVED_BLACKLIST_ITEM_FROM,blacklistName), item, null));
		} finally {
			cfg.log.setNetworkIdle();
		}
	}
	
	/**
	 * @param name the name of the blacklist to remove
	 * @throws RemoteException if the requested operation could not be performed on the
	 * connected peer
	 */
	public void removeBlacklist(String name) throws RemoteException {
		cfg.log.setNetworkBusy();
		try {
			this.getBlacklist().deleteBlacklist(name);
			firePeerEvent(new PeerSettingChangedEvent(SC_REMOVED_BLACKLIST, name, null));
		} finally {
			cfg.log.setNetworkIdle();
		}
	}
	
	/**
	 * @param oldTagName the name of the bookmark-tag to be renamed
	 * @param newTagName the new tag-name
	 * @throws RemoteException if the requested operation could not be performed on the
	 * connected peer
	 */
	public void renameTag(String oldTagName, String newTagName) throws RemoteException {
		cfg.log.setNetworkBusy();
		try {
			getBookmarks().renameTag(oldTagName, newTagName);
			firePeerEvent(new PeerSettingChangedEvent(
					SC_RENAMED_BOOKMARK_TAG,
					oldTagName,
					newTagName));
		} finally { 
			cfg.log.setNetworkIdle();
		}
	}
	
	/**
	 * Send a message to another peer.
	 * 
	 * @see Message
	 * @param message the message object to send; must contain the {@link Message#getRecipientHash()
	 * recipient hash}
	 * @throws RemoteException if the requested operation could not be performed on the
	 * connected peer
	 * @throws IllegalArgumentException if the recipient hash, the subject or the message-body
	 * is <code>null</code>
	 */
	public void sendMessage(Message message) throws RemoteException {
		if (message.getRecipientHash() == null || message.getSubject() == null || message.getBody() == null)
			throw new IllegalArgumentException("No argument may be null");
		cfg.log.setNetworkBusy();
		try {
			getMessages().sendMessage(
					message.getRecipientHash(),
					message.getSubject(),
					message.getBody());
		} finally {
			cfg.log.setNetworkIdle();
		}
	}
	
	/**
	 * Send a message to another peer.
	 * 
	 * @param recipientHash the peer-hash of the receiving YaCy peer
	 * @param subject the headline/subject of the message
	 * @param message the body of the message
	 * @throws RemoteException if the requested operation could not be performed on the
	 * connected peer
	 * @throws IllegalArgumentException if any argument is null
	 */
	public void sendMessage(String recipientHash, String subject, String message)
			throws RemoteException {
		if (recipientHash == null || recipientHash.length() != 12 || subject == null || message == null)
			throw new IllegalArgumentException("No argument may be null");
		cfg.log.setNetworkBusy();
		try {
			getMessages().sendMessage(recipientHash, subject, message);
		} finally {
			cfg.log.setNetworkIdle();
		}
	}
	
	/**
	 * Shuts down the peer.
	 * 
	 * @throws RemoteException if the requested operation could not be performed on the
	 * connected peer
	 * @throws PInterruptionEx if the {@link PeerShuttingDownEvent} has
	 * been interrupted
	 */
	public void shutdownPeer() throws RemoteException, PInterruptionEx {
		firePeerInterruptableEvent(new PeerShuttingDownEvent());
		cfg.log.setNetworkBusy();
		try {
			this.getAdmin().shutdownPeer();
		} finally {
			cfg.log.setNetworkIdle();
		}
	}
	
	/**
	 * @param file the file or folder on the local filesystem to upload
	 * @param workingDirPath the top-folder where the specified file or folder shall be put
	 * @param comment if file is a single file and no directory, a comment may be set
	 * @param indexFile specify whether the file shall be indexed by the peer or not
	 * @return the started Thread performing the upload
	 * @throws IOException if the specified file can not be read
	 */
	public Thread uploadShareFile(
			File file,
			String workingDirPath,
			String comment,
			boolean indexFile) throws IOException {
		if (!file.exists()) throw new IOException("file does not exist: " + file.getPath());
		if (!file.canRead()) throw new IOException("file can not be read: " + file.getPath());
		
		uploadThread d = new uploadThread(file, workingDirPath, comment, indexFile);
		d.start();
		return d;
	}
	
	private class uploadThread extends Thread {
		
		private File file = null;
		private String workingDirPath = null;
		private String comment = null;
		private boolean indexFile = false;
		
		private ShareService service = getShare();
		
		public uploadThread(File file, String workingDirPath, String comment, boolean indexFile) {
			this.file = file;
			this.workingDirPath = workingDirPath;
			this.comment = comment;
			this.indexFile = indexFile;
		}
		
		public void run() {
			long start = System.currentTimeMillis();
			cfg.log.setNetworkBusy();
			
			if (this.file.isDirectory())
				uploadShareFolder(this.file, this.workingDirPath, this.indexFile);
			else
				uploadShareFile(this.file, this.workingDirPath, this.comment, this.indexFile);

			cfg.log.setNetworkIdle();
			cfg.log.addEntry("Übertragung abgeschlossen für " + workingDirPath + "/" + file.getName(), this.getName(), false);
			firePeerEvent(new PeerTransferFinishedEvent(
					file.getPath(),
					workingDirPath + "/" + file.getName(),
					true,
					System.currentTimeMillis() - start,
					null,
					indexFile));
		}
		
		private void uploadShareFolder(File folder, String workingDirPath, boolean indexFiles) {
			try {
				this.service.createDirectory(workingDirPath, folder.getName());
			} catch (RemoteException e) {
				cfg.log.addEntry("Erstellung des Ordners " + folder.getName() + " fehlgeschlagen: " + e.getMessage(), null, false);
				cfg.log(e);
			}
			
			for (File d : folder.listFiles()) {
				if (!d.canRead()) {
					cfg.log(new IOException("Cannot read file " + d));
					continue;
				}
				
				cfg.log.addEntry("Übertrage Datei " + folder.getName() + File.separator + d.getName(), this.getName(), true);
				if (d.isDirectory())
					uploadShareFolder(d, workingDirPath + "/" + folder.getName(), indexFiles);
				else
					uploadShareFile(d, workingDirPath + "/" + folder.getName(), "", indexFiles);
			}
		}
		
		private void uploadShareFile(File file, String workingDirPath, String comment, boolean indexFile) {
			cfg.log.addEntry("Übertrage Datei: " + file.getName(), this.getName(), true);
			((Stub)this.service)._setProperty(Call.ATTACHMENT_ENCAPSULATION_FORMAT, Call.ATTACHMENT_ENCAPSULATION_FORMAT_MIME);
			AttachmentPart part = new AttachmentPart();
			part.setDataHandler(new DataHandler(new FileDataSource(file)));
			part.setContentId(file.getName());
			part.setContentType("application/octet-stream");
			((Stub)this.service).addAttachment(part);
			
			try {
				this.service.uploadFile(workingDirPath, indexFile, comment);
			} catch (RemoteException e) {
				cfg.log(e);
				cfg.log.addEntry("Übertragung fehlgeschlagen: " + e.getMessage(), null, false);
			}
			
			((Stub)this.service).clearAttachments();
		}
	}
	
	/* -----------------------------------------------------------------------
	 * --------------------------- set functions -----------------------------
	 * ----------------------------------------------------------------------- */
	
	/**
	 * @see Peer#blacklistTypes for a list of available services
	 * @param attr the name of the service this blacklist shall be enabled for
	 * @param name the name of the blacklist
	 * @throws RemoteException if the requested operation could not be performed on the
	 * connected peer
	 */
	public void setBlacklistAttribute(String attr, String name) throws RemoteException {
		ArrayList<String> r = this.getBlacklistAttributes(name);
		Iterator<String> it = r.iterator();
		while (it.hasNext())
			if (it.next().equals(attr))
				return;
		r.add(attr);
		cfg.log.setNetworkBusy();
		try {
			this.getBlacklist().activateBlacklist(name, r.toArray(new String[r.size()]));
			// TODO: firePeerEvent(new PeerSettingsChangedEvent())
		} finally {
			cfg.log.setNetworkIdle();
		}
	}
	
	/**
	 * @see Peer#blacklistTypes for a list of available services
	 * @param attr all names of the services this blacklist shall be enabled for
	 * @param name the name of the blacklist
	 * @throws RemoteException if the requested operation could not be performed on the
	 * connected peer
	 */
	public void setBlacklistAttributes(String[] attr, String name) throws RemoteException {
		boolean found;
		for (String d : attr) {
			found = false;
			for (String c : this.blacklistTypes)
				if (d.equals(c)) {
					found = true;
					break;
				}
			if (!found)
				throw new IllegalArgumentException("attr contains illegal blacklist type: " + d);
		}
		cfg.log.setNetworkBusy();
		try {
			this.getBlacklist().activateBlacklist(name, attr);
			// TODO: firePeerEvent(new PeerSettingsChangedEvent())
		} finally {
			cfg.log.setNetworkIdle();
		}
	}
	
	/**
	 * @param yes specify whether the blacklist shall be shared with other peers or not
	 * @param name the name of the blacklist
	 * @throws RemoteException if the requested operation could not be performed on the
	 * connected peer
	 */
	public void setBlacklistShared(boolean yes, String name) throws RemoteException {
		cfg.log.setNetworkBusy();
		try {
			boolean oldValue = isBlacklistShared(name);
			if (yes)
				this.getBlacklist().shareBlacklist(name);
			else
				this.getBlacklist().unshareBlacklist(name);
			firePeerEvent(new PeerSettingChangedEvent(
					String.format(SC_SHARING_BLACKLIST, ((yes) ? ("") : ("un")), name),
					oldValue,
					yes));
		} finally {
			cfg.log.setNetworkIdle();
		}
	}
	
	/**
	 * @param pause set the pause-state of all queues of this peer
	 * @throws RemoteException if the requested operation could not be performed on the
	 * connected peer
	 */
	public void setCrawlStateAll(boolean pause) throws RemoteException {
		cfg.log.setNetworkBusy();
		try {
			boolean oldValue = isCrawlingPaused();
			if (pause)
				getCrawl().pauseCrawling();
			else
				getCrawl().resumeCrawling();
			firePeerEvent(new PeerSettingChangedEvent(SC_QUEUE_STATES_ALL, oldValue, pause));
		} finally {
			cfg.log.setNetworkIdle();
		}
	}
	
	/**
	 * @param localCrawl set the pause-state of the local crawl queue for this peer
	 * @param remoteTriggeredCrawl set the pause-state of the remote triggered crawl queue for
	 * this peer
	 * @param globalCrawlTrigger set the pause-state of the globla crawl trigger queue for this
	 * peer
	 * @throws RemoteException if the requested operation could not be performed on the
	 * connected peer
	 */
	public void setCrawlState(
			boolean localCrawl,
			boolean remoteTriggeredCrawl,
			boolean globalCrawlTrigger) throws RemoteException {
		cfg.log.setNetworkBusy();
		try {
			getCrawl().pauseResumeCrawling(localCrawl, remoteTriggeredCrawl, globalCrawlTrigger);
		} finally {
			cfg.log.setNetworkIdle();
		}
	}
	
	/**
	 * @param enabled enable or disable index distribution to other peers
	 * @throws RemoteException if the requested operation could not be performed on the
	 * connected peer
	 */
	public void setIndexDistribution(boolean enabled) throws RemoteException {
		boolean oldValue = this.indexDistributionEnabled;
		this.indexDistributionEnabled = enabled;
		setTransferProperties();
		firePeerEvent(new PeerSettingChangedEvent(SC_INDEX_DISTRIBUTION, oldValue, enabled));
	}
	
	/**
	 * @param enabled enable or disable index distribution to other peers while crawling
	 * @throws RemoteException if the requested operation could not be performed on the
	 * connected peer
	 */
	public void setIndexDistributionWhileCrawling(boolean enabled) throws RemoteException {
		boolean oldValue = this.indexDistributionWhileCrawling;
		this.indexDistributionWhileCrawling = enabled;
		setTransferProperties();
		firePeerEvent(new PeerSettingChangedEvent(SC_INDEX_DISTRIBUTION_WHILE_CRAWLING, oldValue, enabled));
	}
	
	/**
	 * @param enabled enable or disable index receive from other peers
	 * @throws RemoteException if the requested operation could not be performed on the
	 * connected peer
	 */
	public void setIndexReceive(boolean enabled) throws RemoteException {
		boolean oldValue = this.indexReceiveEnabled;
		this.indexReceiveEnabled = enabled;
		setTransferProperties();
		firePeerEvent(new PeerSettingChangedEvent(SC_INDEX_RECEIVE, oldValue, enabled));
	}
	
	/**
	 * @param enabled enable or disable the check against all active blacklists for received
	 * URLs from other peers
	 * @throws RemoteException if the requested operation could not be performed on the
	 * connected peer
	 */
	public void setIndexReceiveBlockBlacklist(boolean enabled) throws RemoteException {
		boolean oldValue = this.indexReceiveBlockBlacklist;
		this.indexReceiveBlockBlacklist = enabled;
		setTransferProperties();
		firePeerEvent(new PeerSettingChangedEvent(SC_INDEX_RECEIVE_BLOCKED, oldValue, enabled));
	}
	
	/**
	 * @param name change the name of this peer
	 * @throws RemoteException if the requested operation could not be performed on the
	 * connected peer
	 */
	public void setName(String name) throws RemoteException {
		cfg.log.setNetworkBusy();
		try {
			this.getAdmin().setPeerName(name);
			firePeerEvent(new PeerSettingChangedEvent(SC_PEER_NAME, this.name, name));
			this.name = name;
		} finally {
			cfg.log.setNetworkIdle();
		}
	}
	
	/**
	 * @param port change the port this peers listens on for incoming connections
	 * @throws RemoteException if the requested operation could not be performed on the
	 * connected peer
	 */
	public void setPort(int port) throws RemoteException {
		cfg.log.setNetworkBusy();
		try {
			this.getAdmin().setPeerPort(port);
			firePeerEvent(new PeerSettingChangedEvent(SC_PORT, this.port, port));
			this.port = port;
		} finally {
			cfg.log.setNetworkIdle();
		}
	}
	
	/**
	 * @param indexDistributionEnabled enable or disable index distribution to other peers
	 * @param indexDistributionWhileCrawling enable or disable index distribution to other peers
	 * while crawling
	 * @param indexReceiveEnabled enable or disable index receive from other peers
	 * @param indexReceiveBlockBlacklist enable or disable the check against all active
	 * blacklists for received URLs from other peers
	 * @throws RemoteException if the requested operation could not be performed on the
	 * connected peer
	 */
	public void setTransferProperties(
			boolean indexDistributionEnabled,
			boolean indexDistributionWhileCrawling,
			boolean indexReceiveEnabled,
			boolean indexReceiveBlockBlacklist) throws RemoteException {
		cfg.log.setNetworkBusy();
		try {
			getAdmin().setTransferProperties(
					indexDistributionEnabled,
					indexDistributionWhileCrawling,
					indexReceiveEnabled,
					indexReceiveBlockBlacklist);
			firePeerEvent(new PeerSettingChangedEvent(SC_INDEX_TRANSFER,
					new Object[] {
							this.indexDistributionEnabled,
							this.indexDistributionWhileCrawling,
							this.indexReceiveEnabled,
							this.indexReceiveBlockBlacklist },
					new Object[] {
							indexDistributionEnabled,
							indexDistributionWhileCrawling,
							indexReceiveEnabled,
							indexReceiveBlockBlacklist }));
			this.indexDistributionEnabled = indexDistributionEnabled;
			this.indexDistributionWhileCrawling = indexDistributionWhileCrawling;
			this.indexReceiveEnabled = indexReceiveEnabled;
			this.indexReceiveBlockBlacklist = indexReceiveBlockBlacklist;
		} finally {
			cfg.log.setNetworkIdle();
		}
	}
}
