/* 
 * Autor: Franz Brauße
 * License: GPL v2 
 */

package yacy.event;

import yacy.Peer;

/**
 * <p>Thrown when the peer is shut down or the connection is going to be terminated.
 * Cancel this event via it's method <code>cancelEvent</code> to request time to
 * finish your current transaction with the peer.</p>
 * <p>The triggering method will receive a {@link yacy.exceptions.PInterruptionEx}
 * and may then decide whether to override your request or not.</p>
 * <p>{@link PeerInterruptableEvent}s may be interrupted to signalize the originator
 * that more time is needed to fulfill a task. An interruption does not guarrantee
 * that the blocking Thread won't be killed, so advise the thread to finish quickly.</p>
 */
public class PeerInterruptableEvent {
	
	private String 	peerName = null;
	private String 	peerURL = null;
	private Peer 	peer = null;
	private String 	cause = null;
	private Thread 	reason = null;
	
	public Peer 	getPeer() 		{ return this.peer; }
	public String 	getPeerName() 	{ return this.peerName; }
	public String 	getPeerURL() 	{ return this.peerURL; }
	
	/**
	 * {@link PeerInterruptableEvent}s may be interrupted to signalize the originator
	 * that more time is needed to fulfill a task. An interruption does not guarrantee
	 * that the blocking Thread won't be killed, so advise the thread to finish quickly.
	 * 
	 * @param cause a human readable String defining why the event shall be interrupted 
	 * @param reason the blocking Thread because of which this event is being interrupted
	 * @throws IllegalArgumentException if reason is no proper Thread
	 */
	public void interrupt(String cause, Thread reason) throws IllegalArgumentException {
		if (cause == null)
			cause = "Undefined interruption";
		this.cause = cause;
		if (reason == null)
			throw new IllegalArgumentException("A Thread must be passed to cancel a PeerInterruptableEvent!");
		this.reason = reason;
	}
	
	/**
	 * @return whether this event has been interrupted
	 */
	public boolean isInterrupted() {
		return (this.cause != null && this.reason != null);
	}
	
	/**
	 * @return the cause for the interruption
	 */
	public String getInterruptionCause() {
		return this.cause;
	}
	
	/**
	 * @return the blocking thread
	 */
	public Thread getInterruptionReason() {
		return this.reason;
	}
	
	public void setPeer(Peer peer) {
		this.peer = peer;
	}
	public void setPeerName(String peerName) {
		this.peerName = peerName;
	}
	public void setPeerURL(String peerURL) {
		this.peerURL = peerURL;
	}
	
}
