/* 
 * Autor: Franz Brauße
 * License: GPL v2 
 */

package yacy.exceptions;

import yacy.event.PeerInterruptableEvent;

/**
 * <p>A {@link PInterruptionEx} is thrown when a {@link PeerInterruptableEvent}
 * has been {@link PeerInterruptableEvent#interrupt(String, Thread) canceled}.</p>
 * <p>It provides the possibility to terminate the blocking Thread.
 */
public class PInterruptionEx extends PeerException {
	
	private static final long serialVersionUID = 1L;
	
	private static final String exMsg = "Event interrupted: ";
	
	private PeerInterruptableEvent event = null;
	
	/**
	 * <p>A {@link PInterruptionEx} is thrown when a {@link PeerInterruptableEvent}
	 * has been {@link PeerInterruptableEvent#interrupt(String, Thread) canceled}.</p>
	 * <p>It provides the possibility to terminate the blocking Thread.
	 * 
	 * @see PeerInterruptableEvent#interrupt(String, Thread)
	 * @param interruptedEvent the event blocked by a Thread
	 */
	public PInterruptionEx(PeerInterruptableEvent interruptedEvent) {
		super(exMsg + interruptedEvent.toString());
		this.event = interruptedEvent;
	}
	
	/**
	 * @return the original event
	 */
	public PeerInterruptableEvent getInterruptedEvent() { return this.event; }
	
	/**
	 * @return the cause for blocking (mostly the function of the reason-Thread)
	 */
	public String getInterruptionCause() { return this.event.getInterruptionCause(); }
	
	/**
	 * @return the blocking Thread
	 */
	public Thread getInterruptionReason() { return this.event.getInterruptionReason(); }
	
	/**
	 * You may override a requested interruption. The blocking Thread will be forced to
	 * terminate.
	 * 
	 * @see PeerInterruptableEvent#interrupt(String, Thread)
	 * @return <code>true</code> if the Thread has successfully been terminated,
	 * <code>false</code> if an error occured, i.e. because no Thread had been passed to
	 * {@link PeerInterruptableEvent#interrupt(String, Thread)} (this case should not occur).
	 */
	@SuppressWarnings("deprecation")
	public boolean overrideInterruption() {
		if (this.event.getInterruptionReason() == null)
			return false;
		else
			this.event.getInterruptionReason().stop();
		return true;
	}
}
